/*
** ---------------------------------------------------------------------------
** SVN Repository ...: $HeadURL: file:///Data/_Privat/Programmieren/Source-Repository/SVN/BeanDocument/trunk/src/net/beandocument/BeanElement.java $
** Author ...........: $LastChangedBy: jscheffl $
** Modified .........: $LastChangedDate: 2005-03-21 22:42:35 +0100 (Mo, 21 Mrz 2005) $
** Revision .........: $LastChangedRevision: 37 $
** Responsible ......: Jens Scheffler
** Initial version ..: Jens Scheffler
** ---------------------------------------------------------------------------
*/
package net.beandocument;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Base implementation for bean document elements
 * @author Jens Scheffler
 */
public abstract class BeanElement extends BeanBaseNode implements Element {
    
    /** Name of this element in document context */
    private final String tagName;
    
    /** Map for all additional attributes for this bean */
    private Map additionalAttributes;

    /**
     * Constructor for a new Bean Element
     * @param owner Owned document for the new Element
     * @param parent Parent element of this element
     * @param tagName Name of the new element
     * @param bean Bean which is the base for the element
     */
    protected BeanElement(BeanDocument owner, BeanBaseNode parent, String tagName, Object bean) {
        super(owner, parent);

        this.tagName = tagName;

        additionalAttributes = new HashMap();
        addBeanSpecificAttribute("className", bean.getClass().getName());
    }

    /**
     * Retrieves the named attribute value
     * @param name Name of attribute to retrieve
     * @return Value of the attribute or null if not set
     * @see org.w3c.dom.Element#getAttribute(java.lang.String)
     */
    public String getAttribute(String name) {
        Node att = getAttributeNode(name);
        if (att == null) {
            return null;
        }
        return att.getNodeValue();
    }

    /**
     * Retrieves the named attribute object instance
     * @param name Name of attribute to retrieve
     * @return Attribute node or null if not set
     * @see org.w3c.dom.Element#getAttributeNode(java.lang.String)
     */
    public Attr getAttributeNode(String name) {
        return (Attr) getAttributes().getNamedItem(name);
    }

    /**
     * Retrieves the named attribute object instance with the given namespace.
     * Since namespace functionality is not implemented, only when calling with namespace is null a value will be
     * returned.
     * @param namespace Namespace to use for the retrieved node. must be null to retrieve a value.
     * @param name Name of the attribute to retrieve
     * @return Attribute node with the namespace of null or null if any namespace is given
     * @see org.w3c.dom.Element#getAttributeNodeNS(java.lang.String, java.lang.String)
     */
    public Attr getAttributeNodeNS(String namespace, String name) {
        if (namespace != null) { // if namespace is given, return null!
            return null;
        }
        return getAttributeNode(name);
    }

    /**
     * Retrieves the named attribute value with the given namespace.
     * Since namespace functionality is not implemented, only when calling with namespace is null a value will be
     * returned.
     * @param namespace Namespace to use for the retrieved node. must be null to retrieve a value.
     * @param name Name of the attribute to retrieve
     * @return Attribute value with the namespace of null or null if any namespace is given
     * @see org.w3c.dom.Element#getAttributeNS(java.lang.String, java.lang.String)
     */
    public String getAttributeNS(String namespace, String name) {
        Node att = getAttributeNodeNS(namespace, name);
        if (att == null) {
            return null;
        }
        return att.getNodeValue();
    }

    /**
     * Retrieves all attributes of the element
     * @return Attributes of this element
     * @see org.w3c.dom.Node#getAttributes()
     */
    public NamedNodeMap getAttributes() {
        BeanNamedNodeMap result = new BeanNamedNodeMap();
        // create list of attributes during runtime
        Iterator i = additionalAttributes.keySet().iterator();
        BeanAttribute lastAttribute = null;
        while (i.hasNext()) {
            String key = (String) i.next();
            String value = (String) additionalAttributes.get(key);
            BeanAttribute attribute = new BeanAttribute(owner, parent, validName(key), value);
            result.add(key, attribute);
            
            if (lastAttribute != null) {
                lastAttribute.setNextSibling(attribute);
                attribute.setPreviousSibling(lastAttribute);
            }
            lastAttribute = attribute;
        }
        
        return result;
    }

    /**
     * Retrieves all direct child nodes of this element.
     * NOTE: to prevent infinite recursion, only one level of elements is checked! This does not conform with
     * DOM specification but is required to run stable
     * @param name Tag name to search for in the children, "*" means all child's
     * @return List of child elements matching the given name
     * @see org.w3c.dom.Element#getElementsByTagName(java.lang.String)
     */
    public NodeList getElementsByTagName(String name) {
        // if "*" is given as name, return all child's!
        if ("*".equals(name)) {
            return getChildNodes();
        }

        BeanNodeList result = new BeanNodeList();
        NodeList childs = getChildNodes();
        if (childs != null) {
            for (int i = 0; i < childs.getLength(); i++) {
                Node item = childs.item(i);
                if (item.getNodeName().equals(name)) {
                    result.add(item);
                }
            }
        }
        return result;
    }

    /**
     * Retrieves all direct child nodes of this element by given namespace.
     * NOTE: to prevent infinite recursion, only one level of elements is checked! This does not conform with
     * DOM specification but is required to run stable
     * NOTE: since namespace support is not implemented, only calls with empty namespace will return a result
     * @param namespace Namespace to search child nodes for, need to be null or "*" to retrieve a result!
     * @param name Tag name to search for in the children, "*" means all child's
     * @return Empty list of child nodes
     * @see org.w3c.dom.Element#getElementsByTagNameNS(java.lang.String, java.lang.String)
     */
    public NodeList getElementsByTagNameNS(String namespace, String name) {
        if (namespace != null || !"*".equals(namespace)) {
            return new BeanNodeList();
        }

        return getElementsByTagName(name);
    }

    /**
     * Retrieves the first child node from the element
     * @return First child node or null if no child nodes are present
     * @see org.w3c.dom.Node#getFirstChild()
     */
    public Node getFirstChild() {
        NodeList childs = getChildNodes();
        if (childs.getLength() == 0) {
            return null;
        }
        return childs.item(0);
    }

    /**
     * Retrieves the last child node from the element
     * @return Last child node or null if no child nodes are present
     * @see org.w3c.dom.Node#getLastChild()
     */
    public Node getLastChild() {
        NodeList childs = getChildNodes();
        if (childs.getLength() == 0) {
            return null;
        }
        return childs.item(childs.getLength() - 1);
    }

    /**
     * Retrieves the local element name
     * @return Tag name of this element
     * @see org.w3c.dom.Node#getLocalName()
     */
    public String getLocalName() {
        return getTagName();
    }

    /**
     * Gets the element tag name
     * @return Element tag name
     * @see org.w3c.dom.Node#getNodeName()
     */
    public String getNodeName() {
        return getTagName();
    }

    /**
     * Returns the node type value
     * @return Node.ELEMENT_NODE
     * @see org.w3c.dom.Node#getNodeType()
     */
    public short getNodeType() {
        return Node.ELEMENT_NODE;
    }

    /**
     * Retrieves null value as node value like described in specification
     * @return null value
     * @throws org.w3c.dom.DOMException (never throws this exception)
     * @see org.w3c.dom.Node#getNodeValue()
     */
    public String getNodeValue() throws DOMException {
        return null;
    }

    /**
     * Tag name of this element implementation
     * @return Tag name of this element
     * @see org.w3c.dom.Element#getTagName()
     */
    public String getTagName() {
        return tagName;
    }

    /**
     * Checks whether this element contains the attribute with the given name
     * @param name Attribute name to check for
     * @return true if the given attribute exists, false if not
     * @see org.w3c.dom.Element#hasAttribute(java.lang.String)
     */
    public boolean hasAttribute(String name) {
        return (getAttribute(name) != null);
    }

    /**
     * Checks whether this element contains the attribute with the given name
     * @param namespace Namespace to look for
     * @param name Attribute name to check for
     * @return true if the given attribute exists, false if not
     * @see org.w3c.dom.Element#hasAttributeNS(java.lang.String, java.lang.String)
     */
    public boolean hasAttributeNS(String namespace, String name) {
        return (getAttributeNS(namespace, name) != null);
    }

    /**
     * Checks whether this element contains any attribute
     * @return true if any attribute exists, false if not
     * @see org.w3c.dom.Node#hasAttributes()
     */
    public boolean hasAttributes() {
        return (getAttributes().getLength() > 0);
    }

    /**
     * Checks whether this element contains any child nodes
     * @return true if any child node exists, false if not
     * @see org.w3c.dom.Node#hasChildNodes()
     */
    public boolean hasChildNodes() {
        return (getChildNodes().getLength() > 0);
    }

    /**
     * NOTE: This function is not implemented and therefore will just throw an exception!
     * @param name Attribute name to remove - ignored by this implementation
     * @throws org.w3c.dom.DOMException If method is called
     * @see org.w3c.dom.Element#removeAttribute(java.lang.String)
     */
    public void removeAttribute(String name) throws DOMException {
        throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, MODIFICATION_ERROR_MESSAGE);
    }

    /**
     * NOTE: This function is not implemented and therefore will just throw an exception!
     * @param node Attribute to remove - ignored by this implementation
     * @return (This function will never return normally)
     * @throws org.w3c.dom.DOMException If method is called
     * @see org.w3c.dom.Element#removeAttributeNode(org.w3c.dom.Attr)
     */
    public Attr removeAttributeNode(Attr node) throws DOMException {
        throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, MODIFICATION_ERROR_MESSAGE);
    }

    /**
     * NOTE: This function is not implemented and therefore will just throw an exception!
     * @param namespace Namespace of the attribute - ignored by this implementation
     * @param name Attribute name to remove - ignored by this implementation
     * @throws org.w3c.dom.DOMException If method is called
     * @see org.w3c.dom.Element#removeAttributeNS(java.lang.String, java.lang.String)
     */
    public void removeAttributeNS(String namespace, String name) throws DOMException {
        throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, MODIFICATION_ERROR_MESSAGE);
    }

    /**
     * NOTE: This function is not implemented and therefore will just throw an exception!
     * @param name Attribute to modify - ignored by this implementation
     * @param value Attribute value to set - ignored by this implementation
     * @throws org.w3c.dom.DOMException If method is called
     * @see org.w3c.dom.Element#setAttribute(java.lang.String, java.lang.String)
     */
    public void setAttribute(String name, String value) throws DOMException {
        throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, MODIFICATION_ERROR_MESSAGE);
    }

    /**
     * NOTE: This function is not implemented and therefore will just throw an exception!
     * @param node Node to set - ignored by this implementation
     * @return (This function will never return normally)
     * @throws org.w3c.dom.DOMException If method is called
     * @see org.w3c.dom.Element#setAttributeNode(org.w3c.dom.Attr)
     */
    public Attr setAttributeNode(Attr node) throws DOMException {
        throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, MODIFICATION_ERROR_MESSAGE);
    }

    /**
     * NOTE: This function is not implemented and therefore will just throw an exception!
     * @param node Node to set - ignored by this implementation
     * @return (This function will never return normally)
     * @throws org.w3c.dom.DOMException If method is called
     * @see org.w3c.dom.Element#setAttributeNodeNS(org.w3c.dom.Attr)
     */
    public Attr setAttributeNodeNS(Attr node) throws DOMException {
        throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, MODIFICATION_ERROR_MESSAGE);
    }

    /**
     * NOTE: This function is not implemented and therefore will just throw an exception!
     * @param namespace Attribute namespace to modify - ignored by this implementation
     * @param name Attribute to modify - ignored by this implementation
     * @param value Attribute value to set - ignored by this implementation
     * @throws org.w3c.dom.DOMException If method is called
     * @see org.w3c.dom.Element#setAttributeNS(java.lang.String, java.lang.String, java.lang.String)
     */
    public void setAttributeNS(String namespace, String name, String value) throws DOMException {
        throw new DOMException(DOMException.NO_MODIFICATION_ALLOWED_ERR, MODIFICATION_ERROR_MESSAGE);
    }

    /**
     * Sets the node value of this implementation.
     * Since this is defined to be null, calls will be just ignored
     * @param value Value to set
     * @throws org.w3c.dom.DOMException (never throws this exception)
     * @see org.w3c.dom.Node#setNodeValue(java.lang.String)
     */
    public void setNodeValue(String value) throws DOMException {
        // just ignore calls to this method
    }
    
    /**
     * Adds an additional synthetic attribute to the element
     * @param name Name of the attribute to add
     * @param value Value to set to the attribute
     */
    protected void addBeanSpecificAttribute(String name, String value) {
        additionalAttributes.put(name, value);
    }

    /**
     * Validates the name of a element and converts invalid characters for XML result
     * @param name Input name to check
     * @return valid and checked name
     */
    private String validName(String name) {
        String result = name;
        while (result.indexOf("$") != -1) {
            result = result.replace('$', '_');
        }
        return result;
    }
}
