/*
** ---------------------------------------------------------------------------
** SVN Repository ...: $HeadURL: file:///Data/_Privat/Programmieren/Source-Repository/SVN/BeanDocument/trunk/src/net/beandocument/BeanElementFactory.java $
** Author ...........: $LastChangedBy: jscheffl $
** Modified .........: $LastChangedDate: 2005-03-25 14:52:56 +0100 (Fr, 25 Mrz 2005) $
** Revision .........: $LastChangedRevision: 46 $
** Responsible ......: Jens Scheffler
** Initial version ..: Jens Scheffler
** ---------------------------------------------------------------------------
*/
package net.beandocument;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

/**
 * Helper class to analyze a bean for its behavior and data and generate the corresponding element types
 * @author Jens Scheffler
 */
public class BeanElementFactory {
    
    /**
     * Container class for mapping functionality
     * @author jscheffl
     */
    private class Mapping {

        /** Class name which must match */
        private Class match;
        /** Handling element class which will take control for the element */
        private Class handler;

        /**
         * Constructor for a new Mapping instance
         * @param match Matching class
         * @param handler Handling class
         */
        public Mapping(Class match, Class handler) {
            this.match = match;
            this.handler = handler;
        }
        
        /**
         * Get the handler class
         * @return Handler class
         */
        public Class getHandler() {
            return handler;
        }

        /**
         * Gets the matching class
         * @return Matching class
         */
        public Class getMatch() {
            return match;
        }
    }
    
    /** Mapping list for element matching */
    private List elementMappings;
    
    /** Default mapping to use when all other matches did not hit */
    private Mapping defaultMapping;
    
    /** Include filters to apply when evaluating child elements */
    private List includeFilters;

    /** Exclude filters to apply when evaluating child elements */
    private List excludeFilters;
    
    /** owned document for this analyzer */
    private BeanDocument owner;

    /** Maximum evaluation depth for child elements */
    private int maxDepth;

    /**
     * Constructs a new analyzer
     */
    protected BeanElementFactory() {
        super();

        // init the default mappings
        defaultMapping = new Mapping(null, BeanObjectElement.class); 
        elementMappings = new Vector();
        addElementMapping(Collection.class, BeanCollectionElement.class);
        addElementMapping(Enumeration.class, BeanCollectionElement.class);
        addElementMapping(Object[].class, BeanCollectionElement.class);
        addElementMapping(Map.class, BeanMapElement.class);
        addElementMapping(Boolean.class, BeanSimpleTypeElement.class);
        addElementMapping(Number.class, BeanSimpleTypeElement.class);
        addElementMapping(Character.class, BeanSimpleTypeElement.class);
        addElementMapping(Byte[].class, BeanSimpleTypeElement.class);
        addElementMapping(Character[].class, BeanSimpleTypeElement.class);
        addElementMapping(String.class, BeanSimpleTypeElement.class);
        addElementMapping(StringBuffer.class, BeanSimpleTypeElement.class);
    }

    /**
     * Evaluates the provided bean and captures all retrievable element from it as node list containing all elements.
     * @param parent Parent element for the new created element list
     * @param bean Object to evaluate
     * @return NodeList with generated elements from the object properties
     */
    public BeanNodeList evaluateChilds(BeanBaseNode parent, Object bean) {
        BeanNodeList result = new BeanNodeList();
        List rawChilds = new Vector();
        
        // check for null...
        if (bean == null) {
            return result;
        }
        
        // Step 1: Collect all accessible information from the object
        //###########################################################

        // collect all fields
        Field[] fields = bean.getClass().getFields();
        for (int i = 0; i < fields.length; i++) {
            rawChilds.add(fields[i]);
        }
        
        // collect all methods
        Method[] methods = bean.getClass().getMethods();
        for (int i = 0; i < methods.length; i++) {
            Method m = methods[i];
            // filter valid methods
            if (isValidCallMethod(m)) {
                rawChilds.add(m);
            }
        }

        // Step 2: Create the elements based on the findings
        //##################################################
        
        Iterator i = rawChilds.iterator();
        BeanElement lastElement = null;
        while (i.hasNext()) {
            Object item = i.next();
            try {
                String name;
                Object value;

                // fetch details
                if (item instanceof Field) {
                    Field f = (Field) item;
                    name = validName(f.getName());
                    value = f.get(bean);
                } else {
                    Method m = (Method) item;
                    name = validName(m.getName());
                    try {
                        value = m.invoke(bean, null);
                    } catch (Throwable t) {
                        // just ignore calls
                        value = null;
                    }
                }
                
                // create element depending on type
                if (value != null) {
                    // filter the content
                    if (filterOK(parent, bean, name, value)) {
                        
                        // create the new element
                        BeanElement valueElement = getElementForObject(parent, name, value);
                        
                        // add element to result list
                        result.add(valueElement);
                        
                        // link all child elements
                        if (lastElement != null) {
                            valueElement.setPreviousSibling(lastElement);
                            lastElement.setNextSibling(valueElement);
                        }
                        lastElement = valueElement;
                    }
                }
            } catch (Exception e) {
                System.err.println("Error accessing class member \"" + 
                        item + "\" in class " + bean.getClass().getName() + ": " + e);
                e.printStackTrace();
            }
        }
        
        return result;
    }
    
    /**
     * Analyzes the given object and created + returns the correct Element matching to the requested object
     * @param parent Parent node to pass as parent for construction new elements
     * @param name Name for the new Element to create
     * @param bean Object to evaluate and create element from
     * @return Correct BeanElement instance matching to requested parameters
     */
    public BeanElement getElementForObject(BeanBaseNode parent, String name, Object bean) {
        // check all mappings
        Class[] types = new Class[] {BeanDocument.class, BeanBaseNode.class, String.class, Object.class};
        Object[] params = new Object[] {owner, parent, name, bean};
        Iterator i = elementMappings.iterator();
        while (i.hasNext()) {
            Mapping m = (Mapping) i.next();
            if (m.getMatch().isAssignableFrom(bean.getClass())) {
                try {
                    return (BeanElement) m.getHandler().getConstructor(types).newInstance(params);
                } catch (Exception e) {
                    System.err.println("Unable to create element of type " + m.getHandler().getName() + ": " + e);
                    e.printStackTrace();
                }
            }
        }
        // if code execution has come to here, the default mapping will be used...
        try {
            return (BeanElement) defaultMapping.getHandler().getConstructor(types).newInstance(params);
        } catch (Exception e) {
            System.err.println("Unable to create element of type " + defaultMapping.getHandler().getName());
            e.printStackTrace();
        }
        return null;
    }
    
    /**
     * Checks if the given method is valid to call for evaluation
     * @param m Method to evaluate
     * @return flag if calling is safe
     */
    private boolean isValidCallMethod(Method m) {
        // check for number of parameters, need to be null
        if (m.getParameterTypes().length > 0) {
            return false;
        }
        // check if void return type
        if (m.getReturnType().getName().equals("void")) {
            return false;
        }
        // check if class is public
        if (!Modifier.isPublic(m.getDeclaringClass().getModifiers())) {
            return false;
        }
        // check if this would be a constructor call (we don't want to construct new objects!!)
        if (m.getName().equals("newInstance") && m.getDeclaringClass().getName().equals("java.lang.Class")) {
            return false;
        }
        // check if this would be a clone call (we don't want to construct new objects!!)
        if (m.getName().equals("clone")) {
            return false;
        }
        // okay, seems to be fine!
        return true;
    }
    
    /**
     * Adds a handler class to the content element handlers list.
     * @param match Class of which the bean need to be instance of. If null, the default handler will be overridden.
     * @param handler Handler class which will generate the element. This must me subclass of BeanElement
     * @throws IllegalArgumentException if hander is invalid
     */
    public void addElementMapping(Class match, Class handler) throws IllegalArgumentException {
        if (handler == null) {
            throw new IllegalArgumentException("Handler class must not be null!");
        }
        if (!BeanElement.class.isAssignableFrom(handler)) {
            throw new IllegalArgumentException("Handler class \"" + handler.getName() +
                    "\" is not of type BeanElement!");
        }
        if (Modifier.isAbstract(handler.getModifiers())) {
            throw new IllegalArgumentException("Handler class \"" + handler.getName() +
                    "\" must not be abstract!");
        }
        if (match == null) {
            // change the default mapping
            defaultMapping = new Mapping(null, handler);
        } else {
            // add the mapping to the beginning of the list
            List temp = elementMappings;
            elementMappings = new Vector(temp.size() + 1);
            elementMappings.add(new Mapping(match, handler));
            elementMappings.addAll(temp);
        }
    }

    /**
     * Sets the owner for the factory
     * @param owner Owner document for this instance
     */
    protected void setOwner(BeanDocument owner) {
        this.owner = owner;
    }

    /**
     * Sets the maximum evaluation depth for element nodes
     * @param max Maximum depth to evaluate
     */
    public void setMaxDepth(int max) {
        this.maxDepth = max;
    }

    /**
     * Adds an include filter to the set of member inclusion filters
     * @param filter Filter to add to the filter chain
     */
    public void addIncludeFilter(BeanContentFilter filter) {
        if (includeFilters == null) {
            includeFilters = new Vector();
        }
        includeFilters.add(filter);
    }

    /**
     * Adds an exclude filter to the set of member exclusion filters
     * @param filter Filter to add to the filter chain
     */
    public void addExcludeFilter(BeanContentFilter filter) {
        if (excludeFilters == null) {
            excludeFilters = new Vector();
        }
        excludeFilters.add(filter);
    }

    /**
     * Checks if the given element applies to all required filters which were defined for the document.
     * @param parent Parent element for the new content elements
     * @param bean Bean which contains the content which is evaluated
     * @param name Name of the member
     * @param value Value of the member
     * @return Flag if the member applies for all filters and is OK.
     */
    public boolean filterOK(BeanBaseNode parent, Object bean, String name, Object value) {

        // check element depth
        int depth = maxDepth;
        BeanBaseNode n = parent;
        while (n != null) {
            n = n.parent;
            if (depth <= 0) {
                return false;
            }
            depth--;
        }
        
        // check include filters (if available)
        if (includeFilters != null) {
            Iterator i = includeFilters.iterator();
            while (i.hasNext()) {
                BeanContentFilter filter = (BeanContentFilter) i.next();
                if (filter.match(bean.getClass(), name, value.getClass())) {
                    return true;
                }
            }
            return false;
        }
        
        // check exclude filters (if available)
        if (excludeFilters != null) {
            Iterator i = excludeFilters.iterator();
            while (i.hasNext()) {
                BeanContentFilter filter = (BeanContentFilter) i.next();
                if (filter.match(bean.getClass(), name, value.getClass())) {
                    return false;
                }
            }
        }

        return true;
    }

    /**
     * Validates the name of a element and converts invalid characters for XML result
     * @param name Input name to check
     * @return valid and checked name
     */
    private String validName(String name) {
        String result = name;
        while (result.indexOf("$") != -1) {
            result = result.replace('$', '_');
        }
        return result;
    }
}
